@@ -48,6 +48,7 @@ DEF_CRS = "epsg:4326"
 BOX_ORD = "(west, south, east, north)"
 GTYPE = Union[Polygon, MultiPolygon, Tuple[float, float, float, float]]
 GDF = TypeVar("GDF", gpd.GeoDataFrame, gpd.GeoSeries)
+XD = TypeVar("XD", xr.Dataset, xr.DataArray)
 
 __all__ = [
     "snap2nearest",
@@ -56,6 +57,7 @@ __all__ = [
     "arcgis2geojson",
     "geo2polygon",
     "get_transform",
+    "geometry_list",
     "xarray_geomask",
     "gtiff2xarray",
     "xarray2geodf",
@@ -64,6 +66,33 @@ __all__ = [
 ]
 
 
+def arcgis2geojson(
+    arcgis: Union[str, Dict[str, Any]], id_attr: Optional[str] = None
+) -> Dict[str, Any]:
+    """Convert ESRIGeoJSON format to GeoJSON.
+
+    Notes
+    -----
+    Based on `arcgis2geojson <https://github.com/chris48s/arcgis2geojson>`__.
+
+    Parameters
+    ----------
+    arcgis : str or binary
+        The ESRIGeoJSON format str (or binary)
+    id_attr : str, optional
+        ID of the attribute of interest, defaults to ``None``.
+
+    Returns
+    -------
+    dict
+        A GeoJSON file readable by GeoPandas.
+    """
+    if isinstance(arcgis, str):
+        return utils.convert(json.loads(arcgis), id_attr)
+
+    return utils.convert(arcgis, id_attr)
+
+
 def json2geodf(
     content: Union[List[Dict[str, Any]], Dict[str, Any]],
     in_crs: Union[str, pyproj.CRS] = DEF_CRS,
@@ -108,31 +137,101 @@ def json2geodf(
     return geodf
 
 
-def arcgis2geojson(
-    arcgis: Union[str, Dict[str, Any]], id_attr: Optional[str] = None
-) -> Dict[str, Any]:
-    """Convert ESRIGeoJSON format to GeoJSON.
-
-    Notes
-    -----
-    Based on `arcgis2geojson <https://github.com/chris48s/arcgis2geojson>`__.
+def xarray_geomask(
+    ds: XD,
+    geometry: Union[Polygon, MultiPolygon],
+    crs: Union[str, pyproj.CRS],
+    all_touched: bool = False,
+    drop: bool = True,
+    from_disk: bool = False,
+) -> XD:
+    """Mask a ``xarray.Dataset`` based on a geometry.
 
     Parameters
     ----------
-    arcgis : str or binary
-        The ESRIGeoJSON format str (or binary)
-    id_attr : str, optional
-        ID of the attribute of interest, defaults to ``None``.
+    ds : xarray.Dataset or xarray.DataArray
+        The dataset(array) to be masked
+    geometry : Polygon, MultiPolygon
+        The geometry to mask the data
+    crs: str or pyproj.CRS
+        The spatial reference of the input geometry
+    all_touched : bool, optional
+        Include a pixel in the mask if it touches any of the shapes.
+        If False (default), include a pixel only if its center is within one
+        of the shapes, or if it is selected by Bresenham's line algorithm.
+    drop : bool, optional
+        If True, drop the data outside of the extent of the mask geometries.
+        Otherwise, it will return the same raster with the data masked.
+        Default is True.
+    from_disk : bool, optional
+         If True, it will clip from disk using rasterio.mask.mask if possible.
+         This is beneficial when the size of the data is larger than memory.
+         Default is False.
 
     Returns
     -------
-    dict
-        A GeoJSON file readable by GeoPandas.
+    xarray.Dataset or xarray.DataArray
+        The input dataset with a mask applied (np.nan)
     """
-    if isinstance(arcgis, str):
-        return utils.convert(json.loads(arcgis), id_attr)
+    ds_attrs = ds.attrs
+    if isinstance(ds, xr.Dataset):
+        da_attrs = {v: ds[v].attrs for v in ds}
 
-    return utils.convert(arcgis, id_attr)
+    ds = ds.rio.clip([geometry], crs=crs, all_touched=all_touched, drop=drop, from_disk=from_disk)
+    ds.rio.update_attrs(ds_attrs, inplace=True)
+
+    if isinstance(ds, xr.Dataset):
+        _ = [ds[v].rio.update_attrs(da_attrs[v], inplace=True) for v in ds]
+    ds.rio.update_encoding(ds.encoding, inplace=True)
+    return ds
+
+
+def get_gtiff_attrs(
+    resp: bytes,
+    ds_dims: Optional[Tuple[str, str]] = None,
+    driver: Optional[str] = None,
+    nodata: Optional[Number] = None,
+) -> Attrs:
+    """Get nodata, CRS, and dimension names in (vertical, horizontal) order from raster in bytes.
+
+    Parameters
+    ----------
+    resp : bytes
+        Raster response returned from a wed service request such as WMS
+    ds_dims : tuple of str, optional
+        The names of the vertical and horizontal dimensions (in that order)
+        of the target dataset, default to None. If None, dimension names are determined
+        from a list of common names.
+    driver : str, optional
+        A GDAL driver for reading the content, defaults to automatic detection. A list of
+        the drivers can be found here: https://gdal.org/drivers/raster/index.html
+    nodata : float or int, optional
+        The nodata value of the raster, defaults to None, i.e., is determined from the raster.
+
+    Returns
+    -------
+    Attrs
+        No data, CRS, and dimension names for vertical and horizontal directions or
+        a list of the existing dimensions if they are not in a list of common names.
+    """
+    with rio.MemoryFile() as memfile:
+        memfile.write(resp)
+        with memfile.open(driver=driver) as src:
+            r_crs = pyproj.CRS(src.crs)
+            _nodata = utils.get_nodata(src) if nodata is None else nodata
+
+            ds = rxr.open_rasterio(src)
+            if ds_dims is None:
+                ds_dims = utils.get_dim_names(ds)
+
+            valid_dims = list(ds.sizes)
+            if ds_dims is None or any(d not in valid_dims for d in ds_dims):
+                raise MissingAttribute("ds_dims", valid_dims)
+            if isinstance(src.transform, rio.Affine):
+                transform = utils.transform2tuple(src.transform)
+            else:
+                transform = tuple(src.transform)  # type: ignore
+    return Attrs(_nodata, r_crs, ds_dims, transform)
 
 
 def gtiff2xarray(
@@ -142,7 +241,7 @@ def gtiff2xarray(
     ds_dims: Optional[Tuple[str, str]] = None,
     driver: Optional[str] = None,
     all_touched: bool = False,
-    nodata: Union[Number, None] = None,
+    nodata: Optional[Number] = None,
     drop: bool = True,
 ) -> Union[xr.DataArray, xr.Dataset]:
     """Convert (Geo)Tiff byte responses to ``xarray.Dataset``.
@@ -211,6 +310,7 @@ def gtiff2xarray(
                 ds.name = var_name[lyr]
                 dtypes[ds.name] = ds.dtype
                 nodata_dict[ds.name] = utils.get_nodata(vrt) if nodata is None else nodata
+                ds = ds.rio.write_nodata(nodata_dict[ds.name])
                 fpath = Path(tmp_dir, f"{uuid.uuid4().hex}.nc")
                 ds.to_netcdf(fpath)
                 return fpath
@@ -226,15 +326,22 @@ def gtiff2xarray(
         ds[v] = ds[v].astype(dtypes[v])
 
     variables = list(ds)
+    _gm = ds.rio.get_gcps()
+    grid_mapping = _gm.grid_mapping if _gm else "spatial_ref"
+    if grid_mapping in ds:
+        ds = ds.drop_vars(grid_mapping)
     if len(variables) == 1:
         ds = ds[variables[0]].copy()
         ds.attrs["crs"] = attrs.crs.to_string()
         ds.attrs["nodatavals"] = (nodata_dict[ds.name],)
+        ds = ds.rio.write_nodata(nodata_dict[ds.name])
     else:
         ds.attrs["crs"] = attrs.crs.to_string()
         for v in variables:
             ds[v].attrs["crs"] = attrs.crs.to_string()
             ds[v].attrs["nodatavals"] = (nodata_dict[v],)
+            ds[v] = ds[v].rio.write_nodata(nodata_dict[v])
+    ds = ds.rio.write_crs(attrs.crs, grid_mapping_name=grid_mapping)
     if isinstance(geometry, (Polygon, MultiPolygon)):
         if geo_crs is None:
             raise MissingCRS
@@ -242,103 +349,6 @@ def gtiff2xarray(
     return ds
 
 
-def xarray_geomask(
-    ds: Union[xr.Dataset, xr.DataArray],
-    geometry: Union[Polygon, MultiPolygon],
-    crs: Union[str, pyproj.CRS],
-    all_touched: bool = False,
-    drop: bool = True,
-    from_disk: bool = False,
-) -> Union[xr.Dataset, xr.DataArray]:
-    """Mask a ``xarray.Dataset`` based on a geometry.
-
-    Parameters
-    ----------
-    ds : xarray.Dataset or xarray.DataArray
-        The dataset(array) to be masked
-    geometry : Polygon, MultiPolygon
-        The geometry to mask the data
-    crs: str or pyproj.CRS
-        The spatial reference of the input geometry
-    all_touched : bool, optional
-        Include a pixel in the mask if it touches any of the shapes.
-        If False (default), include a pixel only if its center is within one
-        of the shapes, or if it is selected by Bresenham's line algorithm.
-    drop : bool, optional
-        If True, drop the data outside of the extent of the mask geometries.
-        Otherwise, it will return the same raster with the data masked.
-        Default is True.
-    from_disk : bool, optional
-         If True, it will clip from disk using rasterio.mask.mask if possible.
-         This is beneficial when the size of the data is larger than memory.
-         Default is False.
-
-    Returns
-    -------
-    xarray.Dataset or xarray.DataArray
-        The input dataset with a mask applied (np.nan)
-    """
-    ds_attrs = ds.attrs
-    if isinstance(ds, xr.Dataset):
-        da_attrs = {v: ds[v].attrs for v in ds}
-
-    ds = ds.rio.clip([geometry], crs=crs, all_touched=all_touched, drop=drop, from_disk=from_disk)
-    ds.rio.update_attrs(ds_attrs, inplace=True)
-
-    if isinstance(ds, xr.Dataset):
-        _ = [ds[v].rio.update_attrs(da_attrs[v], inplace=True) for v in ds]
-    ds.rio.update_encoding(ds.encoding, inplace=True)
-    return ds
-
-
-def get_gtiff_attrs(
-    resp: bytes,
-    ds_dims: Optional[Tuple[str, str]] = None,
-    driver: Optional[str] = None,
-    nodata: Union[Number, None] = None,
-) -> Attrs:
-    """Get nodata, CRS, and dimension names in (vertical, horizontal) order from raster in bytes.
-
-    Parameters
-    ----------
-    resp : bytes
-        Raster response returned from a wed service request such as WMS
-    ds_dims : tuple of str, optional
-        The names of the vertical and horizontal dimensions (in that order)
-        of the target dataset, default to None. If None, dimension names are determined
-        from a list of common names.
-    driver : str, optional
-        A GDAL driver for reading the content, defaults to automatic detection. A list of
-        the drivers can be found here: https://gdal.org/drivers/raster/index.html
-    nodata : float or int, optional
-        The nodata value of the raster, defaults to None, i.e., is determined from the raster.
-
-    Returns
-    -------
-    Attrs
-        No data, CRS, and dimension names for vertical and horizontal directions or
-        a list of the existing dimensions if they are not in a list of common names.
-    """
-    with rio.MemoryFile() as memfile:
-        memfile.write(resp)
-        with memfile.open(driver=driver) as src:
-            r_crs = pyproj.CRS(src.crs)
-            _nodata = utils.get_nodata(src) if nodata is None else nodata
-
-            ds = rxr.open_rasterio(src)
-            if ds_dims is None:
-                ds_dims = utils.get_dim_names(ds)
-
-            valid_dims = list(ds.sizes)
-            if ds_dims is None or any(d not in valid_dims for d in ds_dims):
-                raise MissingAttribute("ds_dims", valid_dims)
-            if isinstance(src.transform, rio.Affine):
-                transform = utils.transform2tuple(src.transform)
-            else:
-                transform = tuple(src.transform)  # type: ignore
-    return Attrs(_nodata, r_crs, ds_dims, transform)
-
-
 def get_transform(
     ds: Union[xr.Dataset, xr.DataArray],
     ds_dims: Tuple[str, str] = ("y", "x"),
@@ -484,6 +494,26 @@ class Coordinates:
 
     lon: Union[Number, Sequence[Number]]
     lat: Union[Number, Sequence[Number]]
+    bounds: Optional[Tuple[float, float, float, float]] = None
+
+    @staticmethod
+    def __box_geo(bounds: Optional[Tuple[float, float, float, float]]) -> sgeom.Polygon:
+        """Get EPSG:4326 CRS."""
+        if bounds is None:
+            return sgeom.box(*pyproj.CRS(4326).area_of_use.bounds)
+
+        if not isinstance(bounds, (tuple, list)) or len(bounds) != 4:
+            raise InvalidInputType("bounds", "tuple of length 4")
+
+        bbox = sgeom.box(*bounds)
+        if not bbox.within(sgeom.box(*pyproj.CRS(4326).area_of_use.bounds)):
+            raise InvalidInputRange("bounds", "within EPSG:4326")
+        return bbox
+
+    @staticmethod
+    def __validate(pts: gpd.GeoSeries, bbox: sgeom.Polygon) -> gpd.GeoSeries:
+        """Create a ``geopandas.GeoSeries`` from valid coords within a bounding box."""
+        return pts[pts.sindex.query(bbox)].sort_index()
 
     def __post_init__(self) -> None:
         """Normalize the longitude value within [-180, 180)."""
@@ -493,23 +523,8 @@ class Coordinates:
             raise InvalidInputType("lon/lat", "float or list")
 
         lon = np.mod(np.mod(_lon, 360) + 540, 360) - 180
-        pts = gpd.GeoSeries([sgeom.Point(xy) for xy in zip(lon, lat)], crs=self._crs)
-        self._points = self.__validate(pts, self._wgs84)
-
-    @property
-    def _crs(self) -> pyproj.CRS:
-        """Get EPSG:4326 CRS."""
-        return pyproj.CRS(4326)
-
-    @property
-    def _wgs84(self) -> sgeom.Polygon:
-        """Get the WGS84 bounds."""
-        return sgeom.box(*self._crs.area_of_use.bounds)
-
-    @staticmethod
-    def __validate(pts: gpd.GeoSeries, bbox: sgeom.Polygon) -> gpd.GeoSeries:
-        """Create a ``geopandas.GeoSeries`` from valid coords within a bounding box."""
-        return pts[pts.sindex.query(bbox)].sort_index()
+        pts = gpd.GeoSeries([sgeom.Point(xy) for xy in zip(lon, lat)], crs="epsg:4326")
+        self._points = self.__validate(pts, self.__box_geo(self.bounds))
 
     @property
     def points(self) -> gpd.GeoSeries:
@@ -517,6 +532,26 @@ class Coordinates:
         return self._points
 
 
+def validate_crs(val: Union[str, int, pyproj.CRS]) -> str:
+    """Validate a CRS.
+
+    Parameters
+    ----------
+    val : str or int
+        Input CRS.
+
+    Returns
+    -------
+    str
+        Validated CRS as a string.
+    """
+    try:
+        crs: str = pyproj.CRS(val).to_string()
+    except pyproj.exceptions.CRSError as ex:
+        raise InvalidInputType("crs", "a valid CRS") from ex
+    return crs
+
+
 @dataclass
 class Spline:
     """Provide attributes of an interpolated B-spline.
@@ -579,73 +614,6 @@ class GeoBSpline:
     (-97.06127, 32.83319)]
     """
 
-    def __init__(self, points: GDF, npts_sp: int, degree: int = 3) -> None:
-        self.degree = degree
-        self.crs = points.crs
-        if self.crs is None:
-            raise MissingCRS
-
-        if not self.crs.is_projected:
-            raise InvalidInputType("points.crs", "projected CRS")
-
-        if any(points.geom_type != "Point"):
-            raise InvalidInputType("points.geom_type", "Point")
-        self.points = points
-
-        if npts_sp < 1:
-            raise InvalidInputRange("npts_sp", ">= 1")
-        self.npts_sp = npts_sp
-
-        tx, ty = zip(*(g.xy for g in points.geometry))
-        self.x_ln = np.array(tx, dtype="f8").squeeze()
-        self.y_ln = np.array(ty, dtype="f8").squeeze()
-        self.npts_ln = self.x_ln.size
-        self.l_ln = LineString(points.geometry).length
-        self._spline = self.__spline(npts_sp, degree)
-
-    @property
-    def spline(self) -> Spline:
-        """Get the spline as a ``Spline`` object."""
-        return self._spline
-
-    def __spline(self, npts_sp: int, degree: int = 3) -> Spline:
-        """Create a B-spline curve from a set of points.
-
-        Notes
-        -----
-        This function is based on https://stackoverflow.com/a/45928473/5797702.
-
-        Parameters
-        ----------
-        npts_sp : int
-            Number of points in the output spline curve.
-        degree : int, optional
-            Degree of the spline. Should be less than the number of points and
-            greater than 1. Default is 3.
-
-        Returns
-        -------
-        Spline
-            A Spline object with ``x``, ``y``, ``phi``, ``radius``,
-            and ``distance`` attributes.
-        """
-        degree = np.clip(degree, 1, self.npts_ln - 1)
-        konts = np.clip(np.arange(self.npts_ln + degree + 1) - degree, 0, self.npts_ln - degree)
-        spl = BSpline(konts, np.column_stack([self.x_ln, self.y_ln]), degree)
-
-        x_sp, y_sp = spl(np.linspace(0, self.npts_ln - degree, max(npts_sp, 3), endpoint=False)).T
-        phi_sp, rad_sp = self.__curvature(x_sp, y_sp, self.l_ln)
-        geom = (
-            LineString([(x1, y1), (x2, y2)])
-            for x1, y1, x2, y2 in zip(x_sp[:-1], y_sp[:-1], x_sp[1:], y_sp[1:])
-        )
-        d_sp = gpd.GeoSeries(geom, crs=self.crs).length.cumsum().values
-        if npts_sp < 3:
-            idx = np.r_[:npts_sp]
-            return Spline(x_sp[idx], y_sp[idx], phi_sp[idx], rad_sp[idx], d_sp[idx])
-
-        return Spline(x_sp, y_sp, phi_sp, rad_sp, d_sp)
-
     @staticmethod
     def __curvature(
         xs: Union[Sequence[float], np.ndarray], ys: Union[Sequence[float], np.ndarray], l_tot: float
@@ -687,9 +655,76 @@ class GeoBSpline:
         rad[non_small] = scals / dphi[non_small]
         return phi, rad
 
+    def __spline(self, npts_sp: int, degree: int = 3) -> Spline:
+        """Create a B-spline curve from a set of points.
+
+        Notes
+        -----
+        This function is based on https://stackoverflow.com/a/45928473/5797702.
+
+        Parameters
+        ----------
+        npts_sp : int
+            Number of points in the output spline curve.
+        degree : int, optional
+            Degree of the spline. Should be less than the number of points and
+            greater than 1. Default is 3.
+
+        Returns
+        -------
+        Spline
+            A Spline object with ``x``, ``y``, ``phi``, ``radius``,
+            and ``distance`` attributes.
+        """
+        degree = np.clip(degree, 1, self.npts_ln - 1)
+        konts = np.clip(np.arange(self.npts_ln + degree + 1) - degree, 0, self.npts_ln - degree)
+        spl = BSpline(konts, np.column_stack([self.x_ln, self.y_ln]), degree)
+
+        x_sp, y_sp = spl(np.linspace(0, self.npts_ln - degree, max(npts_sp, 3), endpoint=False)).T
+        phi_sp, rad_sp = self.__curvature(x_sp, y_sp, self.l_ln)
+        geom = (
+            LineString([(x1, y1), (x2, y2)])
+            for x1, y1, x2, y2 in zip(x_sp[:-1], y_sp[:-1], x_sp[1:], y_sp[1:])
+        )
+        d_sp = gpd.GeoSeries(geom, crs=self.crs).length.cumsum().values
+        if npts_sp < 3:
+            idx = np.r_[:npts_sp]
+            return Spline(x_sp[idx], y_sp[idx], phi_sp[idx], rad_sp[idx], d_sp[idx])
+
+        return Spline(x_sp, y_sp, phi_sp, rad_sp, d_sp)
+
+    def __init__(self, points: GDF, npts_sp: int, degree: int = 3) -> None:
+        self.degree = degree
+        self.crs = points.crs
+        if self.crs is None:
+            raise MissingCRS
+
+        if not self.crs.is_projected:
+            raise InvalidInputType("points.crs", "projected CRS")
+
+        if any(points.geom_type != "Point"):
+            raise InvalidInputType("points.geom_type", "Point")
+        self.points = points
+
+        if npts_sp < 1:
+            raise InvalidInputRange("npts_sp", ">= 1")
+        self.npts_sp = npts_sp
+
+        tx, ty = zip(*(g.xy for g in points.geometry))
+        self.x_ln = np.array(tx, dtype="f8").squeeze()
+        self.y_ln = np.array(ty, dtype="f8").squeeze()
+        self.npts_ln = self.x_ln.size
+        self.l_ln = LineString(points.geometry).length
+        self._spline = self.__spline(npts_sp, degree)
+
+    @property
+    def spline(self) -> Spline:
+        """Get the spline as a ``Spline`` object."""
+        return self._spline
+
 
 def snap2nearest(lines: GDF, points: GDF, tol: float) -> GDF:
-    """Break lines at specified points at given direction.
+    """Find the nearest points on a line to a set of points.
 
     Parameters
     ----------
@@ -767,6 +802,9 @@ def break_lines(lines: GDF, points: gpd.GeoDataFrame, tol: float = 0.0) -> GDF:
     if (points.direction == "up").sum() + (points.direction == "down").sum() != len(points):
         raise InvalidInputValue("direction", ["up", "down"])
 
+    if not lines.geom_type.isin(["LineString", "MultiLineString"]).all():
+        raise InvalidInputType("geometry", "LineString or MultiLineString")
+
     if lines.crs != points.crs or not lines.crs.is_projected or not points.crs.is_projected:
         crs_proj = "epsg:3857"
         lns = lines.to_crs(crs_proj)
@@ -779,6 +817,11 @@ def break_lines(lines: GDF, points: gpd.GeoDataFrame, tol: float = 0.0) -> GDF:
     if tol > 0.0:
         pts = snap2nearest(lns, pts, tol)
 
+    mlines = lns.geom_type == "MultiLineString"
+    if mlines.any():
+        lns.loc[mlines, "geometry"] = lns.loc[mlines, "geometry"].apply(lambda g: list(g.geoms))
+        lns = lns.explode("geometry").set_crs(crs_proj)
+
     pts_idx, flw_idx = lns.sindex.query_bulk(pts.geometry)
     if len(pts_idx) == 0:
         raise ValueError("No intersection between lines and points")  # noqa: TC003
@@ -795,8 +838,33 @@ def break_lines(lines: GDF, points: gpd.GeoDataFrame, tol: float = 0.0) -> GDF:
         crs=crs_proj,
         index=idx,
     )
+
     if isinstance(lns, gpd.GeoDataFrame):
         lns.loc[idx, "geometry"] = broken_lines
     else:
         lns.loc[idx] = broken_lines
     return lns.to_crs(lines.crs)
+
+
+def geometry_list(
+    geometry: Union[GTYPE, sgeom.Point, sgeom.MultiPoint, sgeom.LineString, sgeom.MultiLineString]
+) -> Union[sgeom.Polygon, sgeom.Point, sgeom.LineString]:
+    """Get a list of polygons, points, and lines from a geometry."""
+    if isinstance(geometry, (sgeom.Polygon, sgeom.LineString, sgeom.Point)):
+        return [geometry]
+
+    if isinstance(geometry, (sgeom.MultiPolygon, sgeom.MultiLineString, sgeom.MultiPoint)):
+        return list(geometry.geoms)
+
+    if isinstance(geometry, (tuple, list)) and len(geometry) == 4:
+        return [sgeom.box(*geometry)]
+    valid_geoms = [
+        "Polygon",
+        "MultiPolygon",
+        "tuple/list of length 4",
+        "Point",
+        "MultiPoint",
+        "LineString",
+        "MultiLineString",
+    ]
+    raise InvalidInputType("geometry", ", ".join(valid_geoms))
