@@ -3,7 +3,7 @@ import io
 
 import geopandas as gpd
 from pygeoogc import ArcGISRESTful, ServiceURL
-from shapely.geometry import LineString, Point, Polygon
+from shapely.geometry import LineString, MultiPolygon, Point, Polygon
 
 import pygeoutils as geoutils
 from pygeoutils import Coordinates, GeoBSpline
@@ -25,6 +25,15 @@ GEO_NAT = Polygon(
 SMALL = 1e-3
 
 
+def test_geom_list():
+    glist = geoutils.geometry_list(GEO_URB)
+    assert len(glist) == 1
+    glist = geoutils.geometry_list(GEO_URB.bounds)
+    assert len(glist) == 1
+    glist = geoutils.geometry_list(MultiPolygon([GEO_URB, GEO_NAT]))
+    assert len(glist) == 2
+
+
 def test_break_line():
     crs_proj = "epsg:3857"
     lines = gpd.GeoSeries([LineString([[0, 0], [2, 2]])], crs=crs_proj)
@@ -91,18 +100,20 @@ def test_json2geodf():
     assert abs(flw.LENGTHKM.sum() - 8.917 * 2) < SMALL
 
 
-def test_gtiff2array(wms_resp):
+def test_gtiff2array(wms_resp, cover_resp):
     canopy_box = geoutils.gtiff2xarray(wms_resp, GEO_NAT.bounds, DEF_CRS)
     canopy = geoutils.gtiff2xarray(wms_resp, GEO_NAT, DEF_CRS, drop=False)
+    cover = geoutils.gtiff2xarray(cover_resp, GEO_NAT, DEF_CRS)
 
     mask = canopy > 60
     vec = geoutils.xarray2geodf(canopy, "float32", mask)
 
-    expected = 72.042
+    expected = 72.2235
     assert (
         abs(canopy_box.mean().values.item() - expected) < SMALL
         and abs(canopy.mean().values.item() - expected) < SMALL
-        and vec.shape[0] == 1043
+        and vec.shape[0] == 1014
+        and cover.rio.nodata == 0
     )
 
 
